#include <stdarg.h>
#ifdef	USE_TCL_STUBS
#	include "tcl.h"
#	ifdef _WIN32
#		define WIN32_LEAN_AND_MEAN
#		include <windows.h>
#		undef WIN32_LEAN_AND_MEAN
#		include <malloc.h>
#	endif
#else
#	include "tclInt.h"
#	include "tclPort.h"
#endif
#include "generate.h"
#ifndef	DLL_VERSION
#	define	DLL_VERSION	"1.0.0"
#endif
#if	TCL_MAJOR_VERSION<8
#	error Tcl distribution is TOO OLD
#elif TCL_MAJOR_VERSION==8 && TCL_MINOR_VERSION==1
#	define Tcl_ObjGetVar2(I,O1,O2,f) Tcl_GetObjVar2(I,Tcl_GetStringFromObj(O1,0),O2?Tcl_GetStringFromObj(O2,0):O2,f)
#	define Tcl_ObjSetVar2(I,O1,O2,N,f) Tcl_SetObjVar2(I,Tcl_GetStringFromObj(O1,0),O2?Tcl_GetStringFromObj(O2,0):O2,N,f)
#else
#endif
#if TCL_MAJOR_VERSION>8 || TCL_MINOR_VERSION>=2
#	define GETBYTES(objPtr,lengthPtr) Tcl_GetByteArrayFromObj(objPtr, lengthPtr)
#	define NEWBYTES(bytes,length) Tcl_NewByteArrayObj(bytes, length)
#	define SETBYTESLEN(objPtr,length) Tcl_SetByteArrayLength( objPtr, length )
#else
#	define GETBYTES(objPtr,lengthPtr) Tcl_GetStringFromObj(objPtr, lengthPtr)
#	define NEWBYTES(bytes,length) Tcl_NewStringObj(bytes, length)
#	define SETBYTESLEN(objPtr,length) Tcl_SetObjLength( objPtr, length )
#endif


#ifdef	__WIN32__
typedef	HINSTANCE	DLLHANDLEt;
#else
#	include <dlfcn.h>
#	ifndef RTLD_NOW
#		define RTLD_NOW 1
#	endif
#	ifndef RTLD_GLOBAL
#		define RTLD_GLOBAL 0
#	endif
typedef	void*	DLLHANDLEt;
static	void*	GetProcAddress( DLLHANDLEt h, char* sym )
{
	void*	p = (void*)dlsym(h, sym);
	if(!p){
		Tcl_DString newName;
		Tcl_DStringInit(&newName);
		Tcl_DStringAppend(&newName, "_", 1);
		Tcl_DStringAppend(&newName, sym, -1);
		p = (void*)dlsym(h, Tcl_DStringValue(&newName));
		Tcl_DStringFree(&newName);
		}
	return p;
}
#endif
#if	(defined(USE_TCLALLOC) && USE_TCLALLOC!=0) || defined(TCL_MEM_DEBUG)
#define SPECIAL_NEWOBJ
static	char*	ckstrdup(CONST char* s)
{
	size_t	len = strlen(s)+1;
	char* r=ckalloc(len);
	memcpy(r,s,len);
	return r;
}
#else
#define ckstrdup(s) strdup(s)
#endif
	


#ifdef	__WIN32__
typedef	struct	DFunc_t		DFunc_t;
#endif
typedef	struct	DFunc_t*	pDFunc_t;
typedef struct	DLL_List {
					char*			name;
    				DLLHANDLEt		handle;
					unsigned		count;
    				struct DLL_List *next;
				} DLL_List;

typedef struct	Interp_List {
					Tcl_Interp*			interp;
    				struct Interp_List *next;
					pCBD_t				CBD_list;
					pDFunc_t			DFunc_list;
				} Interp_List, *pInterp_List;

#define DFUNC_PC	0
#define DFUNC_I		1
#define DFUNC_D		2
#define DFUNC_F		3
#define DFUNC_S		4
#define DFUNC_SC	5
#define DFUNC_B		6
#define DFUNC_VC	7
#define DFUNC_V		8
#define DFUNC_CB	9
#define DFUNC_N		10
#define DFUNC_VCT	11
#define DFUNC_VT	12

typedef	struct	{
				size_t		FA;
				size_t		len;
				char*		b_v;
				} DFBArg_t, *pDFBArg_t;
				
typedef	struct	DFunc_t {
				size_t		nDA;
				size_t		nFA;
				size_t		ntObjs;
				size_t		nbargs;
				size_t		nsObjs;
				size_t		nCB;
				pCBD_t*		CB;
				size_t		FixedArgLen;
				size_t		TotalArgLen;
				size_t		retlen;
				int			escape;
				void*		func;
				pDFBArg_t	bargs;
				size_t*		len;
				Tcl_Obj**	tObjs;
				Tcl_Obj**	sObjs;
				char*		rtk;
				char*		cvals;
				pInterp_List cd;
				Tcl_Command	token;
				Tcl_Obj*	defList;
				pDFunc_t	next;
#ifdef	__WIN32__
				} DFunc_t, *pDFunc_t;
#else
				} DFunc_t;
#endif

/*
 * a general function with arguments at adr of length alen
 * the excess argument radr is used for returning general types
 * of length greater than 8
 */
typedef	void (*pbf_t)(void*);
static	void	b_func(void* radr, size_t alen,void* adr,pbf_t f)
{
	memcpy(alloca(alen),adr,alen);
	f(radr);
}

/*
 * a general void function with arguments at adr of length alen
 */
typedef void (*pvf_t)(void);
static	void v_func(size_t alen,void* adr,pvf_t f)
{
	memcpy(alloca(alen),adr,alen);
	f();
}

/*
 * a general function 1..4 bytes return with arguments at adr of length alen
 */
typedef int (*pif_t)(void);
static	int i_func(size_t alen,void* adr,pif_t f)
{
	memcpy(alloca(alen),adr,alen);
	return f();
}

/*
 * a general function 5..8 bytes return with arguments at adr of length alen
 */
typedef	struct {long a;long b;} long2;
typedef long2 (*pllf_t)(void);
static	long2 ll_func(size_t alen,void* adr,pllf_t f)
{
	memcpy(alloca(alen),adr,alen);
	return f();
}

/*
 * a general function double return with arguments at adr of length alen
 */
typedef double (*pdf_t)(void);
static	double d_func(size_t alen,void* adr,pdf_t f)
{
	memcpy(alloca(alen),adr,alen);
	return f();
}

/*
 * a general function float return with arguments at adr of length alen
 */
typedef float (*pff_t)(void);
static	float f_func(size_t alen,void* adr,pff_t f)
{
	memcpy(alloca(alen),adr,alen);
	return f();
}

static DLL_List		*dll_List = NULL;		/*List of currently loaded DLL's*/
static Interp_List	*interp_List = NULL;	/*List of using interps*/

static	void	StartInterpLine(Tcl_Interp* interp)
{
	if(*Tcl_GetStringResult(interp)){
		Tcl_AppendResult(interp,"\n",NULL);
		}
}

static int dllUsage( Tcl_Interp* interp, Tcl_Obj *CONST objv[], const char* kind )
{
	char*	name = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp,
			name," error:", kind,"\nUse:\n ", name,
			" path name returndef argdef...\n"
			" calls name within dll path\n"
			"  returndef\n"
			"  v void (ie no return)\n"
			"  f float return\n"
			"  d double return\n"
			"  i int4 return\n"
			"  n n byte return use binary scan to decode (0 means void)\n\n"
			"  argdefs\n"
			"  {d doubleval}\n"
			"  {f floatval}\n"
			"  {i int4val}\n"
			"  {b arbitrarystring}\n"
			"  {*[s]n var} indirect nbyte tcl var\n   n=0/omitted=use existing length\n   s form truncate\n",
			"  {s constindirectstring}\n",
			"  {@ callbackdef}\n",
			NULL
			);
	return TCL_ERROR;
}

static	DLL_List* LoadDLL(char *name)
{
    DLLHANDLEt	h;
    DLL_List	*p;

#ifdef	__WIN32__
    h = LoadLibrary(name);
#else
	h = dlopen(name, RTLD_NOW|RTLD_GLOBAL);
#endif
    if (h != NULL) {
		p = (DLL_List*) ckalloc(sizeof(DLL_List)+strlen(name)+1);
		p->handle = h;
		p->next = dll_List;
		strcpy(p->name=((char*)p)+sizeof(DLL_List),name);
		dll_List = p;
    	}
	else {
#if	defined(__WIN32__) && !defined(USE_TCL_STUBS)
		TclWinConvertError(GetLastError());
#else
#endif
		p = NULL;
    	}
    return p;
}

static	DLL_List* Find_DLL(char* name)
{
    DLL_List *p = dll_List;
    while(p && strcmp(name,p->name)) p = p->next;
	return p;
}

static void Unload_DLLs(void)
{
	DLL_List*	p;
    while(dll_List){
		p = dll_List;
		dll_List = dll_List->next;
#ifdef	__WIN32__
		FreeLibrary(p->handle);
#else
		dlclose(p->handle);
#endif
		ckfree((char*)p);
		}
}

static	Interp_List* Find_Interp(Tcl_Interp* interp)
{
    Interp_List *p = interp_List;
    while(p && p->interp!=interp) p = p->next;
	return p;
}

static	void	destroy_CBD( pCBD_t q )
{
	ckfree((char*)q->cb_name);
	ckfree((char*)q->cb_tcl_proc_name);
	ckfree((char*)q->cb_code);
	ckfree((char*)q->cb_argdef);
	ckfree((char*)q);
}

static	void	Delete_Callbacks( pCBD_t p )
{
	while(p){
		pCBD_t	q = p;
		destroy_CBD(q);
		p = p->next;
		}
}

static void dll_deleteProc(ClientData clientData)
{
    Interp_List*	p = interp_List;
    Interp_List*	q = NULL;
    Interp_List*	c = (Interp_List*)clientData;

	while(p && p!=c){
		q = p;
		p = p->next;
		}
	if(p){
		if(q) q->next = p->next;
		else {
			interp_List = p->next;
			if(!interp_List) Unload_DLLs();
			}
		Delete_Callbacks(p->CBD_list);
		ckfree((char*)p);
		}
}

static	void	safe_free( void* p )
{
	if(p){
		ckfree(p);
		}
}

static	void*	safe_alloc( size_t n, size_t s )
{
	n *= s;
	return n?ckalloc(n):NULL;
}

static	void*	final_realloc( void* p, size_t n, size_t s )
{
	if(n) return ckrealloc(p, n*s);
	ckfree(p);
	return NULL;
}
#define FINAL_REALLOC(p,n) p = final_realloc(p,n,sizeof(*p))

/*structure for keeping info about private copies of strings etc.*/
typedef	struct	{
				void**		args;
				int			nargs;
				int			ntruncs;
				Tcl_Obj**	truncs;
				} sArgs_t, *psArgs_t;

static	void Free_sArgs( psArgs_t sArgs )
{
	while(sArgs->nargs--){
		ckfree(sArgs->args[sArgs->nargs]);
		}
	safe_free(sArgs->args);
	safe_free(sArgs->truncs);
}

static	void Perform_Truncations(  psArgs_t sArgs )
{
	while(sArgs->ntruncs--){
		Tcl_Obj*	sObj = sArgs->truncs[sArgs->ntruncs];
		Tcl_SetObjLength(sObj, strlen(Tcl_GetStringFromObj(sObj,0)));
		}
}

static	pCBD_t	Find_Callback(char* name, pInterp_List cd)
{
	pCBD_t	pCBD = cd->CBD_list;
	while(pCBD){
		if(!strcmp(pCBD->cb_name,name)){
			return pCBD;
			}
		pCBD = pCBD->next;
		}
	return 0;
}

static	char*	Parse_Args( pInterp_List cd, pCB_Escape_t escape, Tcl_Interp *interp,
			int objc, Tcl_Obj *CONST objv[], int* parglen, psArgs_t sArgs )
{
	DynBytes_t	ArgDef;
	int			ok = 1;
	Tcl_Obj*	tObj;
	Tcl_Obj*	kObj;
	Tcl_Obj*	vObj;
	Tcl_Obj*	cObj;
	Tcl_Obj*	sObj;
	Tcl_Obj**	values;
	char*		kind;
	int			i_v;
	double		d_v;
	float		f_v;
	pCBD_t		pCBD;
	void*		adr;
	int			len;
	char*		b_v;
	int			i, j, atrunc;

	i = objc-4;
	DynBytes_Init(&ArgDef);
	sArgs->nargs = sArgs->ntruncs = 0;
	sArgs->args = safe_alloc(i,sizeof(char*));
	sArgs->truncs = safe_alloc(i,sizeof(Tcl_Obj*));

	for(i=4;i<objc;i++){
		tObj = objv[i];
		if(Tcl_IsShared(tObj)) tObj = Tcl_DuplicateObj(tObj);
		if(Tcl_ListObjGetElements(interp, tObj, &j, &values )!=TCL_OK || j<2){
			ok = 0;
			}
		else {
			kObj = values[0];
			if(Tcl_IsShared(kObj)) kObj = Tcl_DuplicateObj(kObj);
			kind = Tcl_GetStringFromObj(kObj, 0);
			vObj = values[1];
			if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
			switch(kind[0]){
				case 'd':
				case 'f':
					if(Tcl_GetDoubleFromObj(interp, vObj, &d_v)!=TCL_OK) ok = 0;
					else if(kind[0]=='f'){
						f_v = (float)d_v;
						adr = &f_v;
						len = 4;
						}
					else {
						adr = &d_v;
						len = 8;
						}
					break;
				case 'i':
					if(Tcl_GetIntFromObj(interp, vObj, &i_v)!=TCL_OK) ok = 0;
					else {
						adr = &i_v;
						len = 4;
						}
					break;
				case 'b':
					if(!(b_v=GETBYTES(vObj,&len)) || len<=0 ) ok = 0;
					else adr = b_v;
					break;
				case 's':
					b_v = ckstrdup(GETBYTES(vObj,0));
					adr = &b_v;
					len = 4;
					sArgs->args[sArgs->nargs++] = b_v;
					break;
				case '*':
					atrunc = kind[1]=='s';
					if(sscanf(kind+atrunc+1,"%d", &j)!=1) j = 0;
					cObj = Tcl_ObjGetVar2(interp, vObj, 0, TCL_PARSE_PART1);
					if(!cObj){
						if(!j) ok = 0;
						else {
							sObj = NEWBYTES(0,0);
							SETBYTESLEN( sObj, j );
							cObj = Tcl_ObjSetVar2(interp, vObj, 0, sObj,
								TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1);
							}
						}
					if(atrunc){
						sArgs->truncs[sArgs->ntruncs++] = cObj;
						}
					if(ok){
						b_v = GETBYTES(cObj,0);
						adr = &b_v;
						len = 4;
						}
					break;
				case '@':
					if(!(pCBD = Find_Callback(b_v=Tcl_GetStringFromObj(vObj,0), cd))){
						Tcl_AppendResult(interp,"Missing callback ", b_v, NULL);
						ok = 0;
						}
					else {
						adr = &(pCBD->cb_code);
						len = 4;
						escape->used++;
						pCBD->cb_escape = escape;
						}
					break;
				default:
					if(!sscanf(kind,"%d",&len) || len<=0) ok = 0;
					else if(!(b_v = GETBYTES(vObj,&j)) || j<(int)len) ok = 0;
					else adr = b_v;
					break;
				}
			if(kObj!=values[0]){
				Tcl_DecrRefCount(kObj);
				}
			if(vObj!=values[1]){
				Tcl_DecrRefCount(vObj);
				}
			}
		if(!ok){
			StartInterpLine(interp);
			Tcl_AppendResult(interp,"bad argument ", Tcl_GetStringFromObj(tObj,NULL), "\n", NULL);
			}
		if(tObj!=objv[i]){
			Tcl_DecrRefCount(tObj);
			}
		if(!ok) break;
		DynBytes_Append(&ArgDef,len,adr);
		DynBytes_Align(&ArgDef);
		}
	if(!ok){
		dllUsage(interp, objv, "argument error");
		ckfree(ArgDef.mem);
		Free_sArgs(sArgs);
		return NULL;
		}

	/*to keep everyone happy we need a fake argument array*/
	if(!(i=ArgDef.mempos)) i=4;
	ArgDef.mem = ckrealloc(ArgDef.mem,i);
	*parglen = i;
	return ArgDef.mem;
}

typedef	int	(*uFunc_t)(Tcl_Interp*, Tcl_Obj *CONST objv[],const char*);
static void* FindProcAdr(Tcl_Interp *interp, Tcl_Obj *CONST objv[], uFunc_t ufunc)
{
	char*			path = Tcl_GetStringFromObj(objv[1],NULL);
	char*			name = Tcl_GetStringFromObj(objv[2],NULL);
	DLL_List*		pDLL;
	char*			msg;
	void*			adr;

	if(path[0]=='@'){
		unsigned x;
		if(!sscanf(name,"%u",&x)){
			msg = "invalid offs/index";
			}
		else {
			if(path[1]=='@'){
				if(!sscanf(path+2,"%u",&adr)){
					msg = "invalid @@vec arg";
					}
				else {
					adr = *(((void **)adr)+x);
					return adr;
					}
				}
			else {
				if(!sscanf(path+1,"%u",&adr)){
					msg = "invalid @addr arg";
					}
				else {
					adr = (void *)((unsigned)adr+x);
					return adr;
					}
				}
			}
		}
	else if(!(pDLL=Find_DLL(path)) && !(pDLL=LoadDLL(path))){
#ifndef	__WIN32__
		Tcl_AppendResult(interp,dlerror(),NULL);
#endif
		msg = "can't map the dll";
		}
	else if(!(adr=(void*)GetProcAddress(pDLL->handle,name))) msg = "can't find the routine";
	else return adr;
	ufunc(interp,objv,msg);
	return NULL;
}

/*
 * dll path func returndef argdefs
 * returndef:
 *		v	void
 *		d	a double
 *		f	a float
 * 		i	a 4 byte int
 * 		n	a binary object of length n bytes
 *
 *	argdefs
 *		{b	string}	binary string eg produced by the binary command
 *					length rounded up to alignment 4
 *		{i	value}	4 byte int eg 34 0xfff0ffff
 *		{f	value}	float
 *		{d	value}	double
 *		{s	string}	constant string passed by address
 *		{*n	name}
 *		(*sn name}	pointer to name; a string variable will be produced
 *					of the length n and the pointer to it will be passed
 *					s form will be truncated to first zerobyte
 *
 *		{@	callbackdef} a callback name defined by the ::dll::callback define name
 */
static int dll_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	pInterp_List	cd = (pInterp_List)clientData;
	char*		retdef;
	char*		argdef;
	int			arglen;
	int			retlen;
	void*		adr;
	sArgs_t		sArgs;
	CB_Escape_t	escape;
	int			result;

	Tcl_ResetResult(interp);
	if(objc<4) return dllUsage(interp,objv,"not enough arguments");
	if(!(adr=FindProcAdr(interp,objv,dllUsage))) return TCL_ERROR;
	retdef = Tcl_GetStringFromObj(objv[3],NULL);
	escape.used = 0;
	if(!(argdef = Parse_Args(cd, &escape, interp,objc,objv,&arglen,&sArgs))) return TCL_ERROR;
	if(arglen%4) return dllUsage(interp,objv,"arglen%4 != 0");
	if(escape.used){
		if(setjmp(escape.escape)){
			result = TCL_ERROR;
			goto L_exit;
			}
		}
	result = TCL_OK;
	switch(retdef[0]){
		case 'v':
			v_func(arglen,argdef,(pvf_t)adr);
			break;
		case 'f':
			Tcl_SetObjResult(interp,Tcl_NewDoubleObj((double)f_func(arglen,argdef,(pff_t)adr)));
			break;
		case 'd':
			Tcl_SetObjResult(interp,Tcl_NewDoubleObj(d_func(arglen,argdef,(pdf_t)adr)));
			break;
		case 'i':
			Tcl_SetObjResult(interp,Tcl_NewIntObj(i_func(arglen,argdef,(pif_t)adr)));
			break;
		default:
			if(Tcl_GetIntFromObj(interp,objv[3],&retlen)!=TCL_OK
				|| retlen<0) return dllUsage(interp,objv,"bad returndef");
			if(!retlen){
				v_func(arglen,argdef,(pvf_t)adr);
				}
			else {
				long2 ll2v;
				char* retval;
				if(retlen>8){
					/*paragraph align for safety*/
					size_t	alen = (retlen+15)&0xfffffff0;
					retval = (char*)ckalloc(alen);
					memset(retval,0,alen);
					b_func(retval,arglen,argdef,(pbf_t)adr);
					}
				else {
					retval = (char*)&ll2v;
					if(retlen<=4){
						ll2v.a = 0;
						*((int*)retval) = i_func(arglen,argdef,(pif_t)adr);
						}
					else if(retlen<=8){
						ll2v.b = 0;
						*((long2*)retval) = ll_func(arglen,argdef,(pllf_t)adr);
						}
					}
				Tcl_SetObjResult(interp,Tcl_NewStringObj(retval,retlen));
				if(retval!=(char*)&ll2v) ckfree(retval);
				}
		}

L_exit:
	ckfree(argdef);
	Perform_Truncations(&sArgs);
	Free_sArgs(&sArgs);
	return result;
}

static	int	DFuncInit(pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	Tcl_ResetResult(interp);
	pDFunc->nsObjs = pDFunc->ntObjs = 0;
	if((size_t)objc!=pDFunc->nFA){
		Tcl_AppendResult(interp,Tcl_GetStringFromObj(objv[0],NULL)," not enough arguments",NULL);
		return TCL_ERROR;
		}
	pDFunc->TotalArgLen = pDFunc->FixedArgLen;
	if(pDFunc->nbargs){
		pDFBArg_t	b = pDFunc->bargs;
		pDFBArg_t	blim = b + pDFunc->nbargs;
		int			len;
		while(b<blim){
			Tcl_Obj* vObj = objv[b->FA];
			if(Tcl_IsShared(vObj)){
				vObj = Tcl_DuplicateObj(vObj);
				pDFunc->sObjs[pDFunc->nsObjs++] = vObj;
				}
			if(!(b->b_v = Tcl_GetStringFromObj(vObj,&len)) || len<=0) return TCL_ERROR;
			b->len = len;
			pDFunc->TotalArgLen += (size_t)((len+3)&0xfffffffc);
			b++;
			}
		}
	return TCL_OK;
}

static	void DFuncFinalise(pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	Tcl_Obj**	p=pDFunc->tObjs;
	Tcl_Obj**	pLim=p+pDFunc->ntObjs;
	/*do truncations*/
	while(p<pLim){
		Tcl_SetObjLength(*p, strlen(Tcl_GetStringFromObj(*p,0)));
		p++;
		}
	p = pDFunc->sObjs;
	pLim = p+pDFunc->nsObjs;
	while(p<pLim){
		Tcl_DecrRefCount(*p);
		p++;
		}
}

static	int DFuncGetArgs( pDFunc_t pDFunc, pCB_Escape_t escape, 
							Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[], char* stack )
{
	char*		stackpos = stack;
	size_t		len, i, slen;
	double		d_v;
	char*		b_v;
	char*		cvals = pDFunc->cvals;
	pDFBArg_t	b = pDFunc->bargs;
	Tcl_Obj*	vObj;
	Tcl_Obj*	sObj;
	Tcl_Obj*	cObj;
	Tcl_Obj**	aObj = (Tcl_Obj**)objv+1;
	size_t		ntObjs;
	pCBD_t		pCBD;
	int			atrunc, ok;

	ntObjs = atrunc = 0;
	for(i=0;i<pDFunc->nDA;i++){
		switch(pDFunc->rtk[i]){
			case DFUNC_PC:	/*pure constant*/
				len = *(size_t*)cvals;
				cvals += sizeof(size_t);
				memcpy(stackpos, cvals, len);
				cvals += len;
				break;
			case DFUNC_I:	/*an integer*/
				vObj = *aObj;
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				if(Tcl_GetIntFromObj(interp, vObj, (int*)stackpos)!=TCL_OK) goto L_exit;
				if(vObj!= *aObj++) Tcl_DecrRefCount(vObj);
				len = sizeof(int);
				break;
			case DFUNC_D:	/*a double*/
				vObj = *aObj;
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				if(Tcl_GetDoubleFromObj(interp, vObj, (double*)stackpos)!=TCL_OK) goto L_exit;
				if(vObj!=*aObj++) Tcl_DecrRefCount(vObj);
				len = sizeof(double);
				break;
			case DFUNC_F:	/*a float*/
				vObj = *aObj;
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				if(Tcl_GetDoubleFromObj(interp, vObj, &d_v)!=TCL_OK) goto L_exit;
				if(vObj!= *aObj++) Tcl_DecrRefCount(vObj);
				*(float*)stackpos = (float)d_v;
				len = sizeof(float);
				break;
			case DFUNC_S:	/*a string passed by reference*/
				vObj = *aObj;
				if(Tcl_IsShared(vObj)){
					vObj = Tcl_DuplicateObj(vObj);
					pDFunc->sObjs[pDFunc->nsObjs++] = vObj;
					}
				aObj++;
				*(char**)stackpos = Tcl_GetStringFromObj(vObj,NULL);
				len = sizeof(char*);
				break;
			case DFUNC_SC:	/*a constant string passed by reference*/
				len = *(size_t*)cvals;
				cvals += sizeof(size_t);
				*(char**)stackpos = (char*)cvals;
				cvals += len;
				len = sizeof(char*);
				break;
			case DFUNC_CB:	/*a callback*/
				vObj = *aObj;
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				if(!(pCBD = Find_Callback(b_v=Tcl_GetStringFromObj(vObj,0), pDFunc->cd))){
					Tcl_AppendResult(interp,"Missing/Invalid callback ", b_v, NULL);
					if(vObj!= *aObj) Tcl_DecrRefCount(vObj);
					goto L_exit;
					}
				if(vObj!= *aObj++) Tcl_DecrRefCount(vObj);
				pCBD->cb_escape = escape;
				*(pCBD_t*)stackpos = pCBD;
				len = 4;
				break;
			case DFUNC_VT:
				atrunc = 1;
			case DFUNC_V:
				len = *(size_t*)cvals;
				cvals += sizeof(size_t);
				vObj = *aObj;
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				cObj = Tcl_ObjGetVar2(interp, vObj, 0, TCL_PARSE_PART1);
				ok = 1;
				if(!cObj){
					if(!len) ok = 0;
					else {
						sObj = Tcl_NewStringObj(0,0);
						Tcl_SetObjLength( sObj, len );
						cObj = Tcl_ObjSetVar2(interp, vObj, 0, sObj, TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1);
						}
					}
				if(vObj!= *aObj++) Tcl_DecrRefCount(vObj);
				if(!ok) goto L_exit;
				if(atrunc){
					pDFunc->tObjs[ntObjs++] = cObj;
					atrunc = 0;
					}
				*(char**)stackpos = Tcl_GetStringFromObj(cObj,0);
				len = 4;
				break;
			case DFUNC_VCT:
				atrunc = 1;
			case DFUNC_VC:
				len = *(size_t*)cvals;
				cvals += sizeof(size_t);
				slen = *(size_t*)cvals;
				cvals += sizeof(size_t);
				vObj = Tcl_NewStringObj(cvals,0);
				cvals += slen;
				cObj = Tcl_ObjGetVar2(interp, vObj, 0, TCL_PARSE_PART1);
				ok = 1;
				if(!cObj){
					if(!len) ok = 0;
					else {
						sObj = Tcl_NewStringObj(0,0);
						Tcl_SetObjLength( sObj, len );
						cObj = Tcl_ObjSetVar2(interp, vObj, 0, sObj, TCL_LEAVE_ERR_MSG|TCL_PARSE_PART1);
						}
					}
				Tcl_DecrRefCount(vObj);
				if(!ok) goto L_exit;
				if(atrunc){
					pDFunc->tObjs[ntObjs++] = cObj;
					atrunc = 0;
					}
				*(char**)stackpos = Tcl_GetStringFromObj(cObj,0);
				len = 4;
				break;
			case DFUNC_N:	/*an n byte value*/
				vObj = *aObj;
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				slen = *(size_t*)cvals;
				cvals += sizeof(size_t);
				if(!(b_v = Tcl_GetStringFromObj(vObj,&ok)) || ok<(int)len) ok = 0;
				memcpy(stackpos, b_v, len);
				if(vObj!= *aObj++) Tcl_DecrRefCount(vObj);
				if(!ok) goto L_exit;
				len = (len+3)&0xfffffffc;	/*ensure alignment*/
				break;
			case DFUNC_B:
				len = b->len;
				memcpy(stack,b->b_v,len);
				len = (len+3)&0xfffffffc;	/*ensure alignment*/
				break;
			}
		stackpos += len;
		}
	/*now it's ok for truncations to occur when we finish*/
	pDFunc->ntObjs = ntObjs;
	
	/*setup the escapes for constant callbacks*/
	for(i=0;i<pDFunc->nCB;i++){
		pDFunc->CB[i]->cb_escape = escape;
		}
	return TCL_OK;
L_exit:
	return TCL_ERROR;
}

static	int void_DFuncProc( pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{

	CB_Escape_t	escape;
	int			result = TCL_ERROR;
	if(pDFunc->escape){
		if(setjmp(escape.escape)) goto L_exit;
		}
	if(DFuncInit(pDFunc, interp, objc, objv)!=TCL_OK) goto L_exit;
	{
	char*	stack = alloca(pDFunc->TotalArgLen);
	if(DFuncGetArgs(pDFunc, &escape, interp, objc, objv, stack)!=TCL_OK) goto L_exit;
	((pvf_t)pDFunc->func)();
	}
	result = TCL_OK;
L_exit:
	DFuncFinalise(pDFunc, interp, objc, objv);
	return result;
}
static	int int_DFuncProc( pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{

	CB_Escape_t	escape;
	int			result = TCL_ERROR;
	if(pDFunc->escape){
		if(setjmp(escape.escape)) goto L_exit;
		}
	if(DFuncInit(pDFunc, interp, objc, objv)!=TCL_OK) goto L_exit;
	{
	char*	stack = alloca(pDFunc->TotalArgLen);
	if(DFuncGetArgs(pDFunc, &escape, interp, objc, objv, stack)!=TCL_OK) goto L_exit;
#ifdef SPECIAL_NEWOBJ
	{
	int i = ((pif_t)(pDFunc->func))();
	Tcl_SetObjResult(interp,Tcl_NewIntObj(i));
	}
#else
	Tcl_SetObjResult(interp,Tcl_NewIntObj(((pif_t)(pDFunc->func))()));
#endif
	}
	result = TCL_OK;
L_exit:
	DFuncFinalise(pDFunc, interp, objc, objv);
	return result;
}

static	int double_DFuncProc( pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	CB_Escape_t	escape;
	int			result = TCL_ERROR;
	if(pDFunc->escape){
		if(setjmp(escape.escape)) goto L_exit;
		}
	if(DFuncInit(pDFunc, interp, objc, objv)!=TCL_OK) goto L_exit;
	{
	char*	stack = alloca(pDFunc->TotalArgLen);
	if(DFuncGetArgs(pDFunc, &escape, interp, objc, objv, stack)!=TCL_OK) goto L_exit;
#ifdef SPECIAL_NEWOBJ
	{
	double d = ((pdf_t)(pDFunc->func))();
	Tcl_SetObjResult(interp,Tcl_NewDoubleObj(d));
	}
#else
	Tcl_SetObjResult(interp,Tcl_NewDoubleObj(((pdf_t)(pDFunc->func))()));
#endif
	}
	result = TCL_OK;
L_exit:
	DFuncFinalise(pDFunc, interp, objc, objv);
	return result;
}

static	int float_DFuncProc( pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	CB_Escape_t	escape;
	int			result = TCL_ERROR;
	if(pDFunc->escape){
		if(setjmp(escape.escape)) goto L_exit;
		}
	if(DFuncInit(pDFunc, interp, objc, objv)!=TCL_OK) goto L_exit;
	{
	char*	stack = alloca(pDFunc->TotalArgLen);
	if(DFuncGetArgs(pDFunc, &escape, interp, objc, objv, stack)!=TCL_OK) goto L_exit;
#ifdef SPECIAL_NEWOBJ
	{
	double d = (double)((pff_t)(pDFunc->func))();
	Tcl_SetObjResult(interp,Tcl_NewDoubleObj(d));
	}
#else
	Tcl_SetObjResult(interp,Tcl_NewDoubleObj((double)((pff_t)(pDFunc->func))()));
#endif
	}
	result = TCL_OK;
L_exit:
	DFuncFinalise(pDFunc, interp, objc, objv);
	return result;
}

static	int long2_DFuncProc( pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	CB_Escape_t	escape;
	int			result = TCL_ERROR;
	if(pDFunc->escape){
		if(setjmp(escape.escape)) goto L_exit;
		}
	if(DFuncInit(pDFunc, interp, objc, objv)!=TCL_OK) goto L_exit;
	{
	long2	retval;
	char*	stack = alloca(pDFunc->TotalArgLen);
	if(DFuncGetArgs(pDFunc, &escape, interp, objc, objv, stack)!=TCL_OK) goto L_exit;
	retval = ((pllf_t)(pDFunc->func))();
	Tcl_SetObjResult(interp,Tcl_NewStringObj((char*)&retval,pDFunc->retlen));
	}
	result = TCL_OK;
L_exit:
	DFuncFinalise(pDFunc, interp, objc, objv);
	return result;
}

#if	defined(_MSC_VER) && defined(_WIN32) && defined(_M_IX86)
#	if _MSC_VER>=1200
#		pragma optimize("g",off)
#	endif
#endif
static	int gen_DFuncProc( pDFunc_t pDFunc, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	CB_Escape_t	escape;
	int			result = TCL_ERROR;
	if(pDFunc->escape){
		if(setjmp(escape.escape)) goto L_exit;
		}
	if(DFuncInit(pDFunc, interp, objc, objv)!=TCL_OK) goto L_exit;
	{
	char*	retval = alloca(pDFunc->retlen);
	char*	stack = alloca(pDFunc->TotalArgLen);
	if(DFuncGetArgs(pDFunc, &escape, interp, objc, objv, stack)!=TCL_OK) goto L_exit;
	((pbf_t)(pDFunc->func))(retval);
	Tcl_SetObjResult(interp,Tcl_NewStringObj(retval,pDFunc->retlen));
	}
	result = TCL_OK;
L_exit:
	DFuncFinalise(pDFunc, interp, objc, objv);
	return result;
}
#if	defined(_MSC_VER) && defined(_WIN32) && defined(_M_IX86)
#	if _MSC_VER>=1200
#		pragma optimize("",on)
#	endif
#endif

static int declareUsage( Tcl_Interp* interp, Tcl_Obj *CONST objv[], const char* kind )
{
	char*	name = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp,
			name," error:", kind,"\nUse:\n ", name,
			" path name procname returndef argdef...\n"
			" defines procname as an alias for name within dll path\n",
			NULL
			);
	return TCL_ERROR;
}

static	void	FreeDFunc( pDFunc_t pDFunc )
{
	safe_free(pDFunc->rtk);
	safe_free(pDFunc->tObjs);
	safe_free(pDFunc->sObjs);
	safe_free(pDFunc->cvals);
	safe_free(pDFunc->CB);
	safe_free(pDFunc->bargs);
	if(pDFunc->defList){
		pInterp_List 	I=pDFunc->cd;
		pDFunc_t		p = NULL;
		pDFunc_t		q = I->DFunc_list;

		/*unlink from our per interp list*/
		while(q){
			if(q==pDFunc){
				if(p) p->next = q->next;
				else I->DFunc_list = q->next;
				break;
				}
			p = q;
			q = q->next;
			}

#if	0
		Tcl_ListObjReplace(I->interp,pDFunc->defList,0,pDFunc->nDA+3,0,NULL);
#else
		Tcl_DecrRefCount(pDFunc->defList);
#endif
		}
	safe_free(pDFunc);
}

static	pDFunc_t Parse_DeclareArgs( pInterp_List cd, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
	int			ok = 1;
	Tcl_Obj*	tObj;
	Tcl_Obj*	kObj;
	Tcl_Obj*	vObj;
	Tcl_Obj**	values;
	char*		kind;
	int			i_v;
	double		d_v;
	float		f_v;
	char*		b_v;
	pCBD_t		pCBD;
	void*		adr;
	size_t		len;
	int			i, j, a, hasValue;
	unsigned	atrunc;
	pDFunc_t	pDFunc;
	DynBytes_t	CVals;
	char		k;

	pDFunc = (pDFunc_t)ckalloc(sizeof(*pDFunc));
	memset(pDFunc,0,sizeof(*pDFunc));
	
	DynBytes_Init(&CVals);
	a = objc - 4;
	pDFunc->tObjs = safe_alloc(a,sizeof(*pDFunc->tObjs));
	pDFunc->sObjs = safe_alloc(a,sizeof(*pDFunc->sObjs));
	pDFunc->CB = safe_alloc(a,sizeof(*pDFunc->CB));
	pDFunc->bargs = safe_alloc(a,sizeof(*pDFunc->bargs));
	pDFunc->rtk = safe_alloc(a,sizeof(*pDFunc->rtk));

	for(i=5;i<objc;i++){
		tObj = objv[i];
		if(Tcl_IsShared(tObj)) tObj = Tcl_DuplicateObj(tObj);
		if(Tcl_ListObjGetElements(interp, tObj, &j, &values )!=TCL_OK || j<1 || j>2){
			ok = 0;
			}
		else {
			a = i-5;
			kObj = values[0];
			if(Tcl_IsShared(kObj)) kObj = Tcl_DuplicateObj(kObj);
			kind = Tcl_GetStringFromObj(kObj, 0);
			if((hasValue = (j==2))){
				vObj = values[1];
				if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
				}
			switch(kind[0]){
				case 'd':
				case 'f':
					len = kind[0]=='f'? 4 : 8;
					if(hasValue){
						if(Tcl_GetDoubleFromObj(interp, vObj, &d_v)!=TCL_OK) ok = 0;
						else if(len==4){
							f_v = (float)d_v;
							adr = &f_v;
							}
						else {
							adr = &d_v;
							}
						DynBytes_Append(&CVals,sizeof(len),&len);
						DynBytes_Append(&CVals,len,adr);
						k = DFUNC_PC;
						}
					else k = len==4?DFUNC_F:DFUNC_D;
					break;
				case 'i':
					len = sizeof(int);
					if(hasValue){
						if(Tcl_GetIntFromObj(interp, vObj, &i_v)!=TCL_OK) ok = 0;
						else {
							DynBytes_Append(&CVals,sizeof(len),&len);
							DynBytes_Append(&CVals,len,&i_v);
							k = DFUNC_PC;
							}
						}
					else k= DFUNC_I;
					break;
				case 'b':
					if(hasValue){
						if(!(b_v=Tcl_GetStringFromObj(vObj,&j)) || j<1) ok = 0;
						else {
							len = ((j+3)&0xfffffffc);
							DynBytes_Append(&CVals,sizeof(len),&len);
							DynBytes_Append(&CVals,j,b_v);
							DynBytes_Align(&CVals);
							k = DFUNC_PC;
							}
						}
					else {
						pDFunc->bargs[pDFunc->nbargs++].FA = pDFunc->nFA;
						k = DFUNC_B;
						len = 0;
						}
					break;
				case 's':
					if(hasValue){
						b_v = Tcl_GetStringFromObj(vObj,0);
						j = strlen(b_v)+1;
						len = ((j+3)&0xfffffffc);
						DynBytes_Append(&CVals,sizeof(size_t),&len);
						DynBytes_Append(&CVals,j,b_v);
						DynBytes_Align(&CVals);
						k = DFUNC_SC;
						}
					else {
						pDFunc->nsObjs++;
						k = DFUNC_S;
						}
					len = 4;
					break;
				case '*':
					if((atrunc = kind[1]=='s')){
						pDFunc->ntObjs++;
						}
					len = 4;
					if(sscanf(kind+atrunc+1,"%d", &j)!=1) j = 0;
					DynBytes_Append(&CVals, sizeof(size_t), &j);
					if(j<0) ok = 0;
					else if(hasValue){
						b_v = Tcl_GetStringFromObj(vObj,0);
						j = strlen(b_v)+1;
						len = ((j+3)&0xfffffffc);
						DynBytes_Append(&CVals, sizeof(size_t), &len);
						DynBytes_Append(&CVals, j, b_v);
						DynBytes_Align(&CVals);
						k = atrunc ? DFUNC_VCT : DFUNC_VC;
						}
					else {
						k = atrunc ? DFUNC_VT : DFUNC_V;
						}
					break;
				case '@':
					pDFunc->escape++;
					len = 4;
					if(hasValue){
						if(!(pCBD = Find_Callback(b_v=Tcl_GetStringFromObj(vObj,0), cd))){
							Tcl_AppendResult(interp,"Missing/Invalid callback ", b_v, NULL);
							ok = 0;
							}
						else {
							DynBytes_Append(&CVals,sizeof(len),&len);
							DynBytes_Append(&CVals,sizeof(&(pCBD->cb_code)),&(pCBD->cb_code));
							pDFunc->CB[pDFunc->nCB++] = pCBD;
							k = DFUNC_PC;
							}
						}
					else {
						k = DFUNC_CB;
						}
					break;
				default:
					if(!sscanf(kind,"%d",&j) || j<=0) ok = 0;
					else {
						len = ((j+3)&0xfffffffc);
						if(hasValue){
							if(!(b_v = Tcl_GetStringFromObj(vObj,&j)) || j<(int)len) ok = 0;
							else {
								DynBytes_Append(&CVals,sizeof(len),&len);
								DynBytes_Append(&CVals, j, b_v);
								DynBytes_Align(&CVals);
								k = DFUNC_PC;
								}
							}
						else {
							DynBytes_Append(&CVals, sizeof(int), &j);
							k = DFUNC_N;
							}
						}
					break;
				}
			if(kObj!=values[0]){
				Tcl_DecrRefCount(kObj);
				}
			if(hasValue && vObj!=values[1]){
				Tcl_DecrRefCount(vObj);
				}
			}
		if(!ok){
			StartInterpLine(interp);
			Tcl_AppendResult(interp,"bad argument ", Tcl_GetStringFromObj(tObj,NULL), "\n", NULL);
			}
		if(tObj!=objv[i]){
			Tcl_DecrRefCount(tObj);
			}
		if(!ok) break;
		pDFunc->nDA++;
		if(!hasValue) pDFunc->nFA++;
		pDFunc->FixedArgLen += len;
		pDFunc->rtk[a] = k;
		}

	pDFunc->cvals = CVals.mem;

	if(pDFunc->FixedArgLen%4){
		declareUsage(interp,objv,"arglen%4 != 0");
		ok = 0;
		}

	if(!ok){
		declareUsage(interp, objv, "argument error");
		FreeDFunc(pDFunc);
		return NULL;
		}
	pDFunc->nFA++;

	/*to keep everyone happy we need a fake argument array*/
	FINAL_REALLOC(pDFunc->tObjs, pDFunc->ntObjs);
	FINAL_REALLOC(pDFunc->sObjs, pDFunc->nsObjs);
	FINAL_REALLOC(pDFunc->bargs, pDFunc->nbargs);
	FINAL_REALLOC(pDFunc->CB, pDFunc->nCB);
	FINAL_REALLOC(pDFunc->cvals, CVals.mempos);
	return pDFunc;
}

static void DFunc_deleteProc(pDFunc_t pDFunc)
{
	FreeDFunc(pDFunc);
}

static int declare_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	pInterp_List	cd = (pInterp_List)clientData;
	char*		retdef;
	pDFunc_t	pDFunc;
	void*		adr;

	Tcl_ResetResult(interp);
	if(objc<5) return declareUsage(interp,objv,"not enough arguments");
	if(!(adr=FindProcAdr(interp,objv,declareUsage))) return TCL_ERROR;
	retdef = Tcl_GetStringFromObj(objv[4],NULL);
	if(!(pDFunc=Parse_DeclareArgs(cd,interp,objc,objv))) return TCL_ERROR;
	pDFunc->func = adr;
	pDFunc->cd = cd;
	if(!strcmp(retdef,"v")){
		adr = void_DFuncProc;
		}
	else if(!strcmp(retdef,"f")){
		adr = float_DFuncProc;
		}
	else if(!strcmp(retdef,"d")){
		adr = double_DFuncProc;
		}
	else if(!strcmp(retdef,"i")){
		adr = int_DFuncProc;
		}
	else {
		int	retlen;
		if(Tcl_GetIntFromObj(interp,objv[4],&retlen)!=TCL_OK || retlen<0){
				declareUsage(interp,objv,"bad returndef");
				goto L_exit;
				}
		pDFunc->retlen = retlen;
		if(!retlen){
			adr = void_DFuncProc;
			}
		else if(retlen<=4){
			adr = int_DFuncProc;
			}
		else if(retlen<=8){
			adr = long2_DFuncProc;
			}
		else {
			adr = gen_DFuncProc;
			}
		}
	if((pDFunc->token=Tcl_CreateObjCommand( interp,Tcl_GetStringFromObj(objv[3],NULL),
			(Tcl_ObjCmdProc*)adr,(ClientData)pDFunc,(Tcl_CmdDeleteProc*)DFunc_deleteProc))==NULL) goto L_exit;
	/*
	 * This last is only for dinfo
	 * objv is cmdname path rtn proc ret arg1 arg2...
	 *            0     1    2   3    4
	 */ 
	pDFunc->defList = Tcl_NewListObj(2,objv+1);
	Tcl_ListObjReplace(interp,pDFunc->defList,2,0,objc-4,objv+4);
	pDFunc->next = cd->DFunc_list;
	cd->DFunc_list = pDFunc;
	return TCL_OK;
L_exit:
	FreeDFunc(pDFunc);
	return TCL_ERROR;
}

static int dinfo_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	char*		s;
	char*		pat;
	char*		name;
	Tcl_Obj*	pObj;
	pDFunc_t	p;
	int			lobjc;
	Tcl_Obj**	lobjv;
	Tcl_Obj*	eObj;


	Tcl_ResetResult(interp);
	if(objc!=2) goto L_usage;
	pObj = objv[1];
	if(Tcl_IsShared(pObj)) pObj = Tcl_DuplicateObj(pObj);
	pat = Tcl_GetStringFromObj(pObj,NULL);
	p = ((pInterp_List)clientData)->DFunc_list;
	while(p){
		if(Tcl_StringMatch(name=Tcl_GetCommandName(interp,p->token),pat)){
			Tcl_ListObjGetElements(interp,p->defList,&lobjc,&lobjv);
			eObj = Tcl_NewListObj(2,lobjv);
			Tcl_ListObjAppendElement(interp, eObj, Tcl_NewStringObj(name,-1));
			Tcl_ListObjReplace(interp,eObj,3,0,lobjc-2,lobjv+2);
			Tcl_ListObjAppendElement(interp,Tcl_GetObjResult(interp),eObj);
			}
		p = p->next;
		}
	if(pObj!=objv[1]){
		Tcl_DecrRefCount(pObj);
		}
	return TCL_OK;

L_usage:
	s = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp,s," usage error\nuse ",s," pattern",NULL);
	return TCL_ERROR;
}

static int cbUsage( Tcl_Interp* interp, Tcl_Obj *CONST objv[], const char* kind )
{
	char*	name = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp,
			name," error:", kind,"\nUse:\n\t",
			name, " stdcall name returndef argdefs\n\t",
			name, " cdecl name returndef argdefs\n\t",
			name, " delete pattern\n\t",
			name, " info pattern\n",
			NULL
			);
	return TCL_ERROR;
}

/*no point in making this __stdcall as we're using ...*/
static	 void	__cdecl dispatch(pCBD_t pCBD, void* rAdr, ... )
{
	pCBAD_t		pCBAD = pCBD->cb_argdef;
	char*		cb_tcl_proc_name = pCBD->cb_tcl_proc_name;
	size_t		i;
	size_t		nargs = pCBAD->nargs-1;
	pCBA_t		argdef = pCBAD->argdef;
	char		ret_type = argdef[0].argtype;
	size_t		ret_len = argdef[0].arglen;
	va_list		arg;
	int			objc, result;
	Tcl_Obj**	objv;
	Tcl_Obj*	tObj;
	double		d_v;
	unsigned	i_v;
	char*		s_v;
	char*		errMsg;
	Tcl_Interp*	interp = pCBD->cb_interp;
	Tcl_CmdInfo cb_tcl_proc_info;

	if(!Tcl_GetCommandInfo(interp, cb_tcl_proc_name, &cb_tcl_proc_info)){
		errMsg = "Callback tcl proc not found";
		goto L_panic;
		}
	objv = (Tcl_Obj**)ckalloc((objc=nargs+1)*sizeof(*objv));
	objv[0] = Tcl_NewStringObj(cb_tcl_proc_name,-1);
	Tcl_IncrRefCount(objv[0]);
	va_start(arg,rAdr);
	for(i=1;i<=nargs;i++){
		switch(argdef[i].argtype){
			case 'b':
				objv[i] = Tcl_NewStringObj(va_arg(arg,char*),argdef[i].arglen);
				break;
			case 'd':
				objv[i] = Tcl_NewDoubleObj(*va_arg(arg,double*));
				break;
			case 'f':
				objv[i] = Tcl_NewDoubleObj((double)(*va_arg(arg,float*)));
				break;
			case 'i':
				objv[i] = Tcl_NewIntObj(*va_arg(arg,int*));
				break;
			case 's':
				objv[i] = Tcl_NewStringObj(va_arg(arg,char*),-1);
				break;
			}
		Tcl_IncrRefCount(objv[i]);
		}
	va_end(arg);

	/*here's where we call the tcl proc*/
	result = cb_tcl_proc_info.objProc(cb_tcl_proc_info.objClientData, interp, objc, objv);
	for(i=0;i<(size_t)objc;i++){
		Tcl_DecrRefCount(objv[i]);
		}
	ckfree((char*)objv);
	if(result!=TCL_OK){
		errMsg = "error in Callback tcl proc";
		goto L_panic;
		}
	tObj = Tcl_GetObjResult(interp);
	switch(ret_type){
		case 'v':
			break;
		case 'b':
			i_v = 0;
			s_v = Tcl_GetStringFromObj(tObj,&i_v);
			if(!i_v || !s_v){
				errMsg = "Callback tcl proc returned wrong length";
				goto L_panic;
				}
			if(ret_len<i_v) i_v = ret_len;
			memcpy(rAdr,s_v,i_v);
			break;
		case 'd':
			if(Tcl_GetDoubleFromObj(interp, tObj, (double*)rAdr)!=TCL_OK){
				errMsg = "Callback tcl proc didn't return double";
				goto L_panic;
				}
			break;
		case 'f':
			if(Tcl_GetDoubleFromObj(interp, tObj, &d_v)!=TCL_OK){
				errMsg = "Callback tcl proc didn't return float";
				goto L_panic;
				}
			*((float*)rAdr) = (float)d_v;
			break;
		case 's':
		case 'i':
			if(Tcl_GetIntFromObj(interp, tObj, (int*)rAdr)!=TCL_OK){
				errMsg = "Callback tcl proc didn't return int";
				goto L_panic;
				}
			break;
		}
	Tcl_ResetResult(interp);
	return;

L_panic:
	if(pCBD->cb_escape){
		StartInterpLine(interp);
		Tcl_AppendResult(interp,errMsg,"\nCallback ",pCBD->cb_name," ",pCBD->cb_tcl_proc_name, NULL);
		longjmp(pCBD->cb_escape->escape,1);
		}
}

static	pCBAD_t	Parse_CB_Args( Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
	pCBAD_t		pCBAD;
	int			ok = 1;
	Tcl_Obj*	tObj;
	char*		kind;
	char		k0;
	int			len;
	int			i, j;

	pCBAD = (pCBAD_t)ckalloc(sizeof(*pCBAD)+(objc-4)*sizeof(CBA_t));
	pCBAD->nargs = 0;
	pCBAD->targlen = 0;

	for(i=4;i<objc;i++){
		tObj = objv[i];
		if(Tcl_IsShared(tObj)) tObj = Tcl_DuplicateObj(tObj);
		kind = Tcl_GetStringFromObj(tObj, 0);
		switch(k0=kind[0]){
			case 'd':
				len = 8;
				break;
			case 'f':
			case 'i':
				len = 4;
				break;
			case 'b':
				if(sscanf(kind+1,"%d",&j)!=1||j<=0){
					Tcl_AppendResult(interp,"need length for b callback argument\n",NULL);
					ok=0;
					}
				else {
					len = j;
					if(!len) ok = 0;
					}
				break;
			case 's':
				len = 4;
				break;
			case 'v':
				if(pCBAD->nargs){
					Tcl_AppendResult(interp,"v only for function return not argument\n",NULL);
					ok=0;
					}
				else len = 0;
				break;
			default:
				ok = 0;
				break;
			}
		if(tObj!=objv[i]){
			Tcl_DecrRefCount(tObj);
			}
		if(!ok) break;
		pCBAD->argdef[pCBAD->nargs].argtype = k0;
		pCBAD->argdef[pCBAD->nargs].arglen = len;
		if(pCBAD->nargs++) pCBAD->targlen+=((len+3)&0xfffffffc);
		}
	if(!ok){
		cbUsage(interp, objv, "invalid callback argument");
		ckfree((void*)pCBAD);
		return NULL;
		}
	else {
		pCBAD = (pCBAD_t)ckrealloc((void*)pCBAD,sizeof(*pCBAD)+pCBAD->nargs*sizeof(CBA_t));
		}
	return pCBAD;
}

static	void	delete_matching_cb(char* pat, pInterp_List cd)
{
	pCBD_t		pCBD = cd->CBD_list, qCBD, rCBD = NULL;
	while(pCBD){
		if( Tcl_StringMatch(pCBD->cb_name,pat)){
			qCBD = pCBD;
			pCBD = pCBD->next;
			if(rCBD) rCBD->next = pCBD;
			else cd->CBD_list = pCBD;
			destroy_CBD(qCBD);
			}
		else {
			rCBD = pCBD;
			pCBD = pCBD->next;
			}
		}
}

static	char*	trim_space(char* s)
{
	while(isspace(*s)) s++;
	return s;
}

static int cb_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	char*			cmd;
	char*			pat;
	char*			name;
	pInterp_List	cd = (pInterp_List)clientData;
	pCBD_t			pCBD;
	pCBAD_t			pCBAD;
	Tcl_Obj*		ListObj;
	Tcl_Obj*		tObj;
	size_t			i;
	int				sync, nasync_buf;
	char			buf[20];

	Tcl_ResetResult(interp);
	if(objc<2) goto L_missing_args;
	cmd = Tcl_GetStringFromObj(objv[1],NULL);
	if(!(sync=strncmp("async",cmd,5))){
		cmd = trim_space(cmd + 5);
		if(scanf(cmd,"%d",&nasync_buf)){
			while(isdigit(*cmd)) cmd++;
			cmd = trim_space(cmd + 5);
			}
		else {
			nasync_buf = 1;
			}
		}
	if(!strcmp(cmd,"cdecl")||!strcmp(cmd,"stdcall")){
		int		stdcall = cmd[0]=='s';
		/*::dll::callback define name tclprocname retdef argdefs*/
		if(objc<5) goto L_missing_args;
		pat = Tcl_GetStringFromObj(objv[2],NULL);
		if(!pat || !pat[0]) return cbUsage(interp, objv, "missing/invalid callback name");
		name = Tcl_GetStringFromObj(objv[3],NULL);
		if(!name || !name[0]) return cbUsage(interp, objv, "missing/invalid callback proc");
		if((pCBAD =	Parse_CB_Args( interp, objc, objv))){
			delete_matching_cb( pat, cd );
			pCBD = (pCBD_t)ckalloc(sizeof(*pCBD));
			pCBD->next = cd->CBD_list;
			cd->CBD_list = pCBD;
			strcpy(pCBD->cb_name = ckalloc(strlen(pat)+1), pat);
			strcpy(pCBD->cb_tcl_proc_name = ckalloc(strlen(name)+1), name);
			pCBD->cb_argdef = pCBAD;
			pCBD->cb_stdcall = stdcall;
			pCBD->cb_code = Generate_Callback( pCBD, (void*)dispatch );
			pCBD->cb_escape = NULL;
			pCBD->cb_interp = interp;
			return TCL_OK;
			}
		return TCL_ERROR;
		}
	else if(!strcmp(cmd,"delete")){
		/*::dll::callback delete pattern*/
		if(objc<3) goto L_missing_args;
		else if(objc>3) goto L_too_many_args;
		pat = Tcl_GetStringFromObj(objv[2],NULL);
		if(pat && !*pat) goto L_empty_pat;
		delete_matching_cb( pat, cd );
		return TCL_OK;
		}
	else if(!strcmp(cmd,"info")){
		if(objc<3) goto L_missing_args;
		else if(objc>3) goto L_too_many_args;
		pat = Tcl_GetStringFromObj(objv[2],NULL);
		if(pat && !*pat) goto L_empty_pat;
		pCBD = cd->CBD_list;
		while(pCBD){
			if( Tcl_StringMatch(pCBD->cb_name,pat)){
				ListObj = Tcl_NewListObj(0,NULL);
				Tcl_ListObjAppendElement(interp,ListObj,
						tObj=Tcl_NewStringObj(pCBD->cb_stdcall?"stdcall":"cdecl",-1));
				Tcl_ListObjAppendElement(interp,ListObj,tObj=Tcl_NewStringObj(pCBD->cb_name,-1));
				Tcl_ListObjAppendElement(interp,ListObj,tObj=Tcl_NewStringObj(pCBD->cb_tcl_proc_name,-1));
				for(i=0;i<pCBD->cb_argdef->nargs;i++){
					Tcl_ListObjAppendElement(interp,ListObj,
							tObj=Tcl_NewStringObj(Generate_Argdef(pCBD->cb_argdef->argdef+i,buf),-1));
					}
				tObj = Tcl_GetObjResult(interp);
				Tcl_ListObjAppendElement(interp,tObj,ListObj);
				Tcl_SetObjResult(interp,tObj);
				}
			pCBD = pCBD->next;
			}
		return TCL_OK;
		}
	else {
		return cbUsage(interp, objv, "invalid subcommand");
		}

L_missing_args:
	return cbUsage(interp,objv,"not enough arguments");
L_too_many_args:
	return cbUsage(interp,objv,"too many arguments");
L_empty_pat:
	return cbUsage(interp,objv,"empty pattern");
}

static int int2str_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	char*		s;
	int			i;
	Tcl_Obj*	vObj;
	int			result;

	Tcl_ResetResult(interp);
	if(objc!=2) goto L_usage;
	vObj = objv[1];
	if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
	if((result = Tcl_GetIntFromObj(interp, vObj, &i))==TCL_OK){
		Tcl_SetObjResult(interp,Tcl_NewStringObj((char*)i,-1));
		}
	if(vObj!=objv[1]){
		Tcl_DecrRefCount(vObj);
		}
	return result;

L_usage:
	s = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp,s," usage error\nuse ",s," integer",NULL);
	return TCL_ERROR;
}

/*
 * given an integer (assumed to be an address) work out an address relative to it
 *
 * format of the second argument is
 * <kind> ':' <offset>
 * <offset> --> integer [ | ':' <offset>]
 */
static int setoffs_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	char*		s;
	int			i;
	char*		p;
	Tcl_Obj*	vObj;
	int			result, kind, nchars, retlen;
	double		d_v;

	Tcl_ResetResult(interp);
	if(objc<3||objc>4) goto L_usage;
	vObj = objv[1];
	if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
	result = Tcl_GetIntFromObj(interp, vObj, &i);
	if(vObj!=objv[1]){
		Tcl_DecrRefCount(vObj);
		}
	if(result!=TCL_OK) goto L_usage;
	p = (char*)i;
	vObj = objv[2];
	if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
	s = Tcl_GetStringFromObj(vObj,NULL);
	kind = *s;
	if(kind=='s' || kind=='d' || kind== 'f' || kind == 'i') s++;
	else if(sscanf(s,"%d%n",&retlen,&nchars)==1){
		s += nchars;
		kind = 'n';
		}
	else {
L_bad2: Tcl_AppendResult(interp, "invalid second argument",NULL);
		result = TCL_ERROR;
		}
	while(result==TCL_OK){
		if(*s++!=':' || sscanf(s,"%d%n",&i,&nchars)!=1) goto L_bad2;
		p += i;
		s += nchars;
		if(!*s) break;
		p = *(char**)p;
		}
	if(vObj!=objv[2]){
		Tcl_DecrRefCount(vObj);
		}
	if(result!=TCL_OK) goto L_usage;

	switch(kind){
		case 'i':
			Tcl_SetObjResult(interp,Tcl_NewIntObj(*(int*)p));
			break;
		case 'd':
			Tcl_SetObjResult(interp,Tcl_NewDoubleObj(*(double*)p));
			break;
		case 'f':
			Tcl_SetObjResult(interp,Tcl_NewDoubleObj((double)(*(float*)p)));
			break;
		case 's':
			Tcl_SetObjResult(interp,Tcl_NewStringObj(*(char**)p,-1));
			break;
		case 'n':
			Tcl_SetObjResult(interp,Tcl_NewStringObj(p,retlen));
			break;
		}
	if(objc==4){
		Tcl_Obj* rObj = Tcl_GetObjResult(interp);
		Tcl_IncrRefCount(rObj);
		Tcl_ResetResult(interp);
		vObj = objv[3];
		if(Tcl_IsShared(vObj)) vObj = Tcl_DuplicateObj(vObj);
		switch(kind){
			case 'i':
				if(Tcl_GetIntFromObj(interp,vObj,(int*)p)!=TCL_OK){
L_bad3: 			Tcl_AppendResult(interp, "invalid third argument",NULL);
					}
				break;
			case 'd':
				if(Tcl_GetDoubleFromObj(interp,vObj,(double*)p)!=TCL_OK) goto L_bad3;
				break;
			case 'f':
				if(Tcl_GetDoubleFromObj(interp,vObj,&d_v)!=TCL_OK) goto L_bad3;
				*(float*)p = (float)d_v;
				break;
			case 's':
				s = Tcl_GetStringFromObj(vObj,NULL);
				if(s==NULL) goto L_bad3;
				strcpy(*(char**)p,s);
				break;
			case 'n':
				s = Tcl_GetStringFromObj(vObj,&i);
				if(i<retlen||s==NULL) goto L_bad3;
				memcpy(p,s,retlen);
				break;
			}
		if(vObj!=objv[3]){
			Tcl_DecrRefCount(vObj);
			}
		if(result==TCL_OK){
			Tcl_SetObjResult(interp,rObj);
			}
		Tcl_DecrRefCount(rObj);
		}
	if(result!=TCL_OK) goto L_usage;
	return result;

L_usage:
	s = Tcl_GetStringFromObj(objv[0],NULL);
	Tcl_AppendResult(interp,s," usage error\nuse ",s," integer indirection",NULL);
	return TCL_ERROR;
}

/*returns the calling Tcl_Interp as an integer*/
static int get_interp_proc( ClientData clientData, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[] )
{
	if(objc!=1){
		char* s = Tcl_GetStringFromObj(objv[0],NULL);
		Tcl_ResetResult(interp);
		Tcl_AppendResult(interp,s," usage error\nuse ",s,NULL);
		return TCL_ERROR;
		}
	Tcl_SetObjResult(interp,Tcl_NewIntObj((int)interp));
	return TCL_OK;
}

DLLEXPORT int Dll_Init(Tcl_Interp* interp)
{
	Interp_List*	p;
	static	char* ns_cmd=
		"namespace eval ::dll {namespace export call* declare int2str dinfo setoffs;variable library {" DLL_LIBRARY "}\n\
	proc makeAbsPath {path} {\n\
		switch -exact [file pathtype $path] {\n\
			absolute {\n\
				return $path\n\
				}\n\
			relative {\n\
				return [file join [pwd] $path]\n\
				}\n\
			volumerelative {\n\
				set splitpath [file split $path]\n\
				set volume [lindex $splitpath 0]\n\
				set cwd [pwd]\n\
				cd $volume\n\
			    eval set abspath \\[file join [pwd] [lrange $splitpath 1 end]\\]\n\
				cd $cwd\n\
				return $abspath\n\
				}\n\
			}\n\
		}\n\
	}";

#if	defined(USE_TCL_STUBS)
	if(Tcl_InitStubs(interp,"8.2.0",0)==NULL) goto L_fail;
#endif
		
	if(!Find_Interp(interp)){
		p = (Interp_List*)ckalloc(sizeof(Interp_List));
		p->interp = interp;
		p->CBD_list = NULL;
		p->DFunc_list = NULL;
		p->next = interp_List;
		interp_List = p;
		if(Tcl_Eval(interp,ns_cmd)==TCL_ERROR){
L_fail:
			Tcl_AppendResult(interp,"\nCan't initialise package dll",NULL);
			interp_List = interp_List->next;
			ckfree((char*)p);
			return TCL_ERROR;
			}
		if(Tcl_CreateObjCommand( interp,(char*)"::dll::call",dll_proc,(ClientData)interp_List,dll_deleteProc)==NULL
			|| Tcl_CreateObjCommand( interp,(char*)"::dll::callback",cb_proc,(ClientData)interp_List,NULL)==NULL
			|| Tcl_CreateObjCommand( interp,(char*)"::dll::int2str",int2str_proc,(ClientData)interp_List,NULL)==NULL
			|| Tcl_CreateObjCommand( interp,(char*)"::dll::declare",declare_proc,(ClientData)interp_List,NULL)==NULL
			|| Tcl_CreateObjCommand( interp,(char*)"::dll::setoffs",setoffs_proc,(ClientData)interp_List,NULL)==NULL
			|| Tcl_CreateObjCommand( interp,(char*)"::dll::dinfo",dinfo_proc,(ClientData)interp_List,NULL)==NULL
			|| Tcl_CreateObjCommand( interp,(char*)"::dll::get_interp",get_interp_proc,(ClientData)NULL,NULL)==NULL
		 	|| Tcl_PkgProvide(interp,"dll",DLL_VERSION)==TCL_ERROR) goto L_fail;
		}
	Tcl_AppendResult(interp,DLL_VERSION,NULL);
	return TCL_OK;
}
