set ::ITERS 0
proc	do_tests {} {
	set ntests 0
	set nfailed 0
	proc print_verbose {test_name test_description contents_of_test code answer} {
		puts stdout "\n"; flush stdout
		puts stdout "==== $test_name $test_description"; flush stdout
		puts stdout "==== Contents of test case:"; flush stdout
		puts stdout "$contents_of_test"; flush stdout
		if {$code != 0} {
		if {$code == 1} {
			puts stdout "==== Test generated error:"; flush stdout
			puts stdout $answer
		} elseif {$code == 2} {
			puts stdout "==== Test generated return exception;  result was:"; flush stdout
			puts stdout $answer; flush stdout
		} elseif {$code == 3} {
			puts stdout "==== Test generated break exception"; flush stdout
		} elseif {$code == 4} {
			puts stdout "==== Test generated continue exception"; flush stdout
		} else {
			puts stdout "==== Test generated exception $code;  message was:"; flush stdout
			puts stdout $answer; flush stdout
		}
		} else {
		puts stdout "==== Result was:"; flush stdout
		puts stdout "$answer"; flush stdout
		}
	}

	proc test {test_name test_description contents_of_test passing_results} {
		set code [catch {uplevel $contents_of_test} answer]
		uplevel incr ntests
		if {$code != 0} {
			uplevel incr nfailed
			print_verbose $test_name $test_description $contents_of_test $code $answer
			} \
		elseif {[string match $passing_results $answer] != 1} { 
			print_verbose $test_name $test_description $contents_of_test $code $answer 
			puts stdout "---- Result should have been:"; flush stdout
			puts stdout "$passing_results"; flush stdout
			puts stdout "---- $test_name FAILED" ; flush stdout
			uplevel incr nfailed
			} \
		elseif {[info exists ::VERBOSE]} {
			print_verbose $test_name $test_description $contents_of_test $code $answer
			}
		}

	proc	locate_dll {pat args} {
		foreach p $args {
			set fn [glob -nocomplain [file join $p $pat]]
			if {"$fn"!={}} break
			}
		return $fn
		}
	global tcl_platform
	set is_win32 [string match $tcl_platform(platform) windows]
	if {$is_win32} {
		set fn [locate_dll dll*.dll ./Release ./objs . .. /tcl/bin]
		load $fn dll
		set ::dll::testdll [locate_dll testdll.dll ./objs . ..]
		} \
	else {
		set fn [locate_dll libdll*.so ./unix . .. /usr/local/lib /tcl/bin]
		load $fn dll
		set ::dll::testdll [locate_dll testdll.so ./unix . ..]
		}
	test 0.0 {call setrecord} {::dll::call $::dll::testdll setrecord v {i 0x1234567}} {}
	test 0.01 {call getrecord} {::dll::call $::dll::testdll getrecord i} {19088743}
	test 0.011 {call setrecord b arg} {::dll::call $::dll::testdll setrecord v [list b [binary format i 0x7abcdef]]} {}
	test 0.012 {call getrecord} {::dll::call $::dll::testdll getrecord i} {128699887}
	test 0.013 {call setrecord 4 arg} {::dll::call $::dll::testdll setrecord v [list  4 [binary format i 0xabcdef9]]} {}
	test 0.014 {call getrecord} {::dll::call $::dll::testdll getrecord i} {180150009}
	test 0.02 {declare getrecord} {::dll::declare $::dll::testdll getrecord getrecord i} {}
	test 0.021 {dinfo getrecord} {::dll::dinfo getrecord} "{$::dll::testdll getrecord getrecord i}"
	test 0.022 {dinfo get*} {::dll::dinfo get*} "{$::dll::testdll getrecord getrecord i}"
	test 0.023 {dinfo *record} {::dll::dinfo *record} "{$::dll::testdll getrecord getrecord i}"
	test 0.024 {dinfo *} {::dll::dinfo *} "{$::dll::testdll getrecord getrecord i}"
	test 0.21 {call setrecord} {::dll::call $::dll::testdll setrecord v {i 12345678}} {}
	test 0.03 {use declared getrecord} {getrecord} {12345678}
	test 0.04 {declare setrecord} {::dll::declare $::dll::testdll setrecord setrecord v i} {}
	test 0.041 {dinfo setrecord} {::dll::dinfo setrecord} "{$::dll::testdll setrecord setrecord v i}"
	test 0.042 {dinfo set*} {::dll::dinfo set*} "{$::dll::testdll setrecord setrecord v i}"
	test 0.043 {dinfo *record} {lsort [::dll::dinfo *record]} [list "$::dll::testdll getrecord getrecord i" "$::dll::testdll setrecord setrecord v i"]
	test 0.044 {dinfo *} {lsort [::dll::dinfo *]} [list "$::dll::testdll getrecord getrecord i" "$::dll::testdll setrecord setrecord v i"]
	test 0.05 {use declared setrecord} {setrecord 0x7654321} {}
	test 0.06 {check declared setrecord} {getrecord} {124076833}
	test 1.0 {call i x i ==> v} {set i [::dll::call $::dll::testdll func_v v {i 7} {i 5} {i 3}]} {}
	test 1.01 {check 1.0} {getrecord} {753}
	test 1.02 {call i x i ==> 0bytes} {::dll::call $::dll::testdll func_v 0 {i 9} {i 2} {i 7}} {}
	test 1.03 {check 1.02} {getrecord} {927}
	test 1.1 {call i x i ==> i} {::dll::call $::dll::testdll func_i i {i 2} {i 3}} {23}
	test 1.2 {call i ==> d} {::dll::call $::dll::testdll func_r d {i 2}} {2.2}
	test 1.3 {call i ==> i} {::dll::call $::dll::testdll func_l4 i {i 10}} {230}
	test 1.4 {call i ==> 8bytes} {binary scan [::dll::call $::dll::testdll func_l8 8 {i 10}] i2 v1;set v1} {11 230}
	test 1.5 {call i ==> 12bytes} {binary scan [::dll::call $::dll::testdll func_l12 12 {i 10}] di v1 v2;list $v1 $v2} {11.0 230}
	test 1.6 {call s ==> i strlen} {::dll::call $::dll::testdll func_i_sa i {s 0123456789}} {10}
	test 2.0 {call __stdcall i x i ==> v} {::dll::call $::dll::testdll sfunc_v v {i 11} {i 8} {i 2}} {}
	test 2.01 {check 2.0} {getrecord} {1182}
	test 2.02 {call __stdcall i x i ==> 0bytes} {::dll::call $::dll::testdll sfunc_v 0 {i 13} {i 6} {i 1}} {}
	test 2.01 {check 2.0} {getrecord} {1361}
	test 2.1 {call __stdcall i x i ==> i} {::dll::call $::dll::testdll sfunc_i i {i 2} {i 3}} {23}
	test 2.2 {call __stdcall i ==> d} {::dll::call $::dll::testdll sfunc_r d {i 2}} {2.2}
	test 2.3 {call __stdcall i ==> i} {::dll::call $::dll::testdll sfunc_l4 i {i 10}} {230}
	test 2.4 {call __stdcall i ==> 8bytes} {binary scan [::dll::call $::dll::testdll sfunc_l8 8 {i 10}] i2 v1;set v1} {11 230}
	test 2.5 {call __stdcall i ==> 12bytes} {binary scan [::dll::call $::dll::testdll sfunc_l12 12 {i 10}] di v1 v2;list $v1 $v2} {11.0 230}
	test 2.6 {call __stdcall s ==> i strlen} {::dll::call $::dll::testdll sfunc_i_sa i {s 0123456789}} {10}
	proc tclCB2 {b} {
		set i [string length $b]
		set j [expr {$i-1}]
		upvar CBV z
		set z [list $i [string index $b 0] [string index $b $j]]
		}
	test 3.0 {callback define} {::dll::callback cdecl CB2 tclCB2 v b277} {}
	test 3.1 {callback info} {::dll::callback info CB2} {{cdecl CB2 tclCB2 v b277}}
	test 3.2 {callback use} {set v [::dll::call $::dll::testdll UseA_cdecl_CB2 v {i 33} {@ CB2}];list $v $CBV} {{} {277 a b}}
	proc testCB1 {name} {
		proc tclCB1 { i s } {
			return [expr {100+$i}]
			}
		::dll::callback cdecl CB1 tclCB1 i i s
		set i 0
		for {set j 0} {$j<10} {incr j} {
			set i [expr {$i+[::dll::call $::dll::testdll $name i [list i $j] {s AString} {@ CB1}]}]
			}
		::dll::callback delete CB1
		rename tclCB1 {}
		return $i
		}
	test 3.3 {callback use} {testCB1 UseA_cdecl_CB1} {1045}
	test 3.31 {callback use} {testCB1 UseA_stdcall_CB1} {1045}

	proc declare_and_call {name retdef args} {
		set decllist "$::dll::testdll $name $name $retdef $args"
		eval ::dll::declare $decllist
		set r [$name]
		set di [::dll::dinfo $name]
		if {[string compare $di "{[string trim $decllist]}"]} {
			append r " dinfo error {$di} != {$decllist}"
			}
		rename $name {}
		return $r
		}
	test 4.0 {declared constant call i x i ==> v} {declare_and_call func_v v {i 15} {i 1} {i 7}} {}
	test 4.01 {check 4.0} {getrecord} {1517}
	test 4.02 {call constant i x i ==> 0bytes} {declare_and_call func_v 0 {i 19} {i 2} {i 0}} {}
	test 4.03 {check 4.02} {getrecord} {1920}
	test 4.1 {call constant i x i ==> i} {declare_and_call func_i i {i 2} {i 3}} {23}
	test 4.2 {call constant i ==> d} {declare_and_call func_r d {i 2}} {2.2}
	test 4.3 {call constant i ==> i} {declare_and_call func_l4 i {i 10}} {230}
	test 4.4 {call constant i ==> 8bytes} {binary scan [declare_and_call func_l8 8 {i 10}] i2 v1;set v1} {11 230}
	test 4.5 {call constant i ==> 12bytes} {binary scan [declare_and_call func_l12 12 {i 10}] di v1 v2;list $v1 $v2} {11.0 230}
	test 4.6 {call constant s ==> i strlen} {declare_and_call func_i_sa i {s 0123456789}} {10}
	proc declare_and_call {name retdef args} {
		set declare "$::dll::testdll $name $name $retdef"
		set call $name
		foreach a $args {
			append declare " [lindex $a 0]"
			append call " [lindex $a 1]"
			}
		eval ::dll::declare $declare
		eval set r \[$call\]
		set di [::dll::dinfo $name]
		if {[string compare $di "{[string trim $declare]}"]} {
			append r " dinfo error {$di} != {$declare}"
			}
		rename $name {}
		return $r
		}
	test 5.0 {declared call i x i ==> v} {declare_and_call func_v v {i 1} {i 2} {i 3}} {}
	test 5.01 {check 5.0} {getrecord} {123}
	test 5.02 {declared call i x i ==> 0bytes} {declare_and_call func_v 0 {i 3} {i 2} {i 1}} {}
	test 5.03 {check 5.02} {getrecord} {321}
	test 5.1 {declared call i x i ==> i} {declare_and_call func_i i {i 2} {i 3}} {23}
	test 5.2 {declared call i ==> d} {declare_and_call func_r d {i 2}} {2.2}
	test 5.3 {declared call i ==> i} {declare_and_call func_l4 i {i 10}} {230}
	test 5.4 {declared call i ==> 8bytes} {binary scan [declare_and_call func_l8 8 {i 10}] i2 v1;set v1} {11 230}
	test 5.5 {declared call i ==> 12bytes} {binary scan [declare_and_call func_l12 12 {i 10}] di v1 v2;list $v1 $v2} {11.0 230}
	test 5.6 {declared call s ==> i strlen} {declare_and_call func_i_sa i {s 0123456789}} {10}
	set ::dll::tAdr0 [::dll::call $::dll::testdll dll_getaddress i {i 0}]
	set ::dll::tAdr1 [::dll::call $::dll::testdll dll_getaddress i {i 1}]

	#two args
	test 6.00 {setoffs s:0}			{::dll::setoffs $::dll::tAdr0 s:0}		{this is level0 string}
	test 6.01 {setoffs i:4}			{::dll::setoffs $::dll::tAdr0 i:4}		{87654321}
	test 6.02 {setoffs d:8}			{::dll::setoffs $::dll::tAdr0 d:8}		{123456.78}
	test 6.03 {setoffs f:16}		{::dll::setoffs $::dll::tAdr0 f:16}		{1234567.0}
	test 6.04 {setoffs 7:20}		{::dll::setoffs $::dll::tAdr0 7:20}		{gfedcba}
	test 6.05 {setoffs s:28:0}		{::dll::setoffs $::dll::tAdr0 s:28:0}	{This is level0 string}
	test 6.06 {setoffs i:32:0}		{::dll::setoffs $::dll::tAdr0 i:32:0}	{12345678}
	test 6.07 {setoffs d:36:0}		{::dll::setoffs $::dll::tAdr0 d:36:0}	{87.654321}
	test 6.08 {setoffs f:40:0}		{::dll::setoffs $::dll::tAdr0 f:40:0}	{7654321.0}
	test 6.09 {setoffs 7:44:0}		{::dll::setoffs $::dll::tAdr0 7:44:0}	{abcdefg}
	test 6.10 {setoffs s:0:0}		{::dll::setoffs $::dll::tAdr1 s:0:0}	{this is level0 string}
	test 6.11 {setoffs i:0:4}		{::dll::setoffs $::dll::tAdr1 i:0:4}	{87654321}
	test 6.12 {setoffs d:0:8}		{::dll::setoffs $::dll::tAdr1 d:0:8}	{123456.78}
	test 6.13 {setoffs f:0:16}		{::dll::setoffs $::dll::tAdr1 f:0:16}	{1234567.0}
	test 6.14 {setoffs 7:0:20}		{::dll::setoffs $::dll::tAdr1 7:0:20}	{gfedcba}
	test 6.15 {setoffs s:0:28:0}	{::dll::setoffs $::dll::tAdr1 s:0:28:0}	{This is level0 string}
	test 6.16 {setoffs i:0:32:0}	{::dll::setoffs $::dll::tAdr1 i:0:32:0}	{12345678}
	test 6.17 {setoffs d:0:36:0}	{::dll::setoffs $::dll::tAdr1 d:0:36:0}	{87.654321}
	test 6.18 {setoffs f:0:40:0}	{::dll::setoffs $::dll::tAdr1 f:0:40:0}	{7654321.0}
	test 6.19 {setoffs 7:0:44:0}	{::dll::setoffs $::dll::tAdr1 7:0:44:0}	{abcdefg}
	test 6.20 {setoffs s:4:0}		{::dll::setoffs $::dll::tAdr1 s:4:0}	{this is level0 string}
	test 6.21 {setoffs i:4:4}		{::dll::setoffs $::dll::tAdr1 i:4:4}	{87654321}
	test 6.22 {setoffs d:4:8}		{::dll::setoffs $::dll::tAdr1 d:4:8}	{123456.78}
	test 6.23 {setoffs f:4:16}		{::dll::setoffs $::dll::tAdr1 f:4:16}	{1234567.0}
	test 6.24 {setoffs 7:4:20}		{::dll::setoffs $::dll::tAdr1 7:4:20}	{gfedcba}
	test 6.25 {setoffs s:4:28:0}	{::dll::setoffs $::dll::tAdr1 s:4:28:0}	{This is level0 string}
	test 6.26 {setoffs i:4:32:0}	{::dll::setoffs $::dll::tAdr1 i:4:32:0}	{12345678}
	test 6.27 {setoffs d:4:36:0}	{::dll::setoffs $::dll::tAdr1 d:4:36:0}	{87.654321}
	test 6.28 {setoffs f:4:40:0}	{::dll::setoffs $::dll::tAdr1 f:4:40:0}	{7654321.0}
	test 6.29 {setoffs 7:4:44:0}	{::dll::setoffs $::dll::tAdr1 7:4:44:0}	{abcdefg}

	#three args
	test 6.30 {setoffs s:4:28:0 v}	{::dll::setoffs $::dll::tAdr1 s:4:28:0 Jessikat}				{This is level0 string}
	test 6.31 {setoffs i:4:32:0 v}	{::dll::setoffs $::dll::tAdr1 i:4:32:0 87654321}				{12345678}
	test 6.32 {setoffs d:4:36:0 v}	{::dll::setoffs $::dll::tAdr1 d:4:36:0 123456.78}				{87.654321}
	test 6.33 {setoffs f:4:40:0 v}	{::dll::setoffs $::dll::tAdr1 f:4:40:0 1234567.0}				{7654321.0}
	test 6.34 {setoffs 7:4:44:0 v}	{::dll::setoffs $::dll::tAdr1 7:4:44:0 gfedcba}					{abcdefg}
	test 6.35 {setoffs s:4:28:0 v}	{::dll::setoffs $::dll::tAdr1 s:4:28:0 {This is level0 string}}	{Jessikat}
	test 6.36 {setoffs i:4:32:0 v}	{::dll::setoffs $::dll::tAdr1 i:4:32:0 12345678}				{87654321}
	test 6.37 {setoffs d:4:36:0 v}	{::dll::setoffs $::dll::tAdr1 d:4:36:0 87.654321}				{123456.78}
	test 6.38 {setoffs f:4:40:0 v}	{::dll::setoffs $::dll::tAdr1 f:4:40:0 7654321.0}				{1234567.0}
	test 6.39 {setoffs 7:4:44:0 v}	{::dll::setoffs $::dll::tAdr1 7:4:44:0 abcdefg}					{gfedcba}
	if {![string compare $::tcl_platform(platform) windows]} {
		proc DavidKarpa { i } {
			::dll::declare $::dll::testdll rtx_StartTxLabel RTX_StartTxLabel i i i i i
			set oi [getrecord]
			setrecord $i
			set result [RTX_StartTxLabel 0 200 354778 8774]
			setrecord $oi
			rename RTX_StartTxLabel {}
			return $result
			}
		test 13.0 {DavidKarpa} {DavidKarpa 0} {0}
		test 13.1 {DavidKarpa} {DavidKarpa 1} {200}
		test 13.2 {DavidKarpa} {DavidKarpa 2} {0}
		test 13.3 {DavidKarpa} {DavidKarpa 3} {27098}
		test 13.4 {DavidKarpa} {DavidKarpa 4} {5}
		test 13.5 {DavidKarpa} {DavidKarpa 5} {8774}

		catch {rename DavidKarpa {}}
		}
	proc adeclare_and_call {adr offs name retdef args} {
		set decllist "$adr $offs $name $retdef $args"
		eval ::dll::declare $decllist
		set r [$name]
		set di [::dll::dinfo $name]
		if {[string compare $di "{[string trim $decllist]}"]} {
			append r " dinfo error $di != {$decllist}"
			}
		rename $name {}
		return $r
		}
	set ::dll::tAdr0 [::dll::call $::dll::testdll dll_getaddress i {i 3}]
	set ::dll::tAdr1 [::dll::call $::dll::testdll dll_getaddress i {i 4}]
	test 7.00 {call @setrecord} {::dll::call @$::dll::tAdr0 0 v {i 0x1234567}} {}
	test 7.01 {call @getrecord} {::dll::call @$::dll::tAdr1 0 i} {19088743}
	test 7.02 {call @setrecord} {::dll::call @$::dll::tAdr0 0 v {i 0}} {}
	test 7.03 {call @getrecord} {::dll::call @$::dll::tAdr1 0 i} {0}
	test 7.04 {declare call @setrecord} {adeclare_and_call @$::dll::tAdr0 0 srec v {i 0x1234567}} {}
	test 7.05 {declare call @getrecord} {adeclare_and_call @$::dll::tAdr1 0 grec i} {19088743}
	set ::dll::tAdr0 [::dll::call $::dll::testdll dll_getaddress i {i 2}]
	test 7.10 {call @@vec 0} {::dll::call @@$::dll::tAdr0 0 i {i 1}} {2}
	test 7.11 {call @@vec 1} {::dll::call @@$::dll::tAdr0 1 i {i 2}} {7}
	test 7.12 {call @@vec 2} {::dll::call @@$::dll::tAdr0 2 i {i 3}} {14}
	test 7.20 {declared call @@vec 0} {adeclare_and_call @@$::dll::tAdr0 0 srec i {i 1}} {2}
	test 7.21 {declared call @@vec 1} {adeclare_and_call @@$::dll::tAdr0 1 srec i {i 2}} {7}
	test 7.22 {declared call @@vec 2} {adeclare_and_call @@$::dll::tAdr0 2 srec i {i 3}} {14}

	test 14.00 {testGetUserName 256}	{}					{}
	catch {::dll::callback delete CB2}
	catch {unset ::dll::testdll}
	catch {unset ::dll::tAdr0}
	catch {unset ::dll::tAdr1}
	catch {rename test {}}
	catch {rename adeclare_and_call {}}
	catch {rename srec {}}
	catch {rename grec {}}
	catch {rename getrecord {}}
	catch {rename setrecord {}}
	catch {rename locate_dll {}}
	catch {rename declare_and_call {}}
	catch {rename print_verbose {}}
	catch {rename tclCB2 {}}
	catch {rename testCB1 {}}
	if {!$nfailed} { set nfailed "no failures" } else {set nfailed "$nfailed failed"}
	return "$ntests tests carried out, $nfailed."
	}
if {[info command memory]!={}} {
	for {set TestIter 0} {$TestIter<30} {incr TestIter} {
		memory tag Iteration$TestIter
		puts "\nIteration $TestIter"
		if {[info exists DLL_INTERP]} {
			interp create test_interp
			interp eval test_interp "proc do_tests {} {[info body do_tests]}"
			interp eval test_interp {
				puts "[do_tests]"
				}
			interp delete test_interp
			} \
		else {
			puts "[do_tests]"
			}
		memory info
		}
	memory active /tmp/fff0
	} \
else {
	if {[info exists ::env(WISHDUMP)]} {
		set f [open $::env(WISHDUMP) w]
		} \
	else {
		set f stdout
		}
	if {$::ITERS>0} {
		puts $f "[time do_tests $::ITERS]"
		}
	puts $f "[do_tests]"
	if {"$f"!="stdout"} {
		close $f
		}
	unset f
	}
